/* -*- Mode:C++; c-file-style:"gnu"; indent-tabs-mode:nil; -*- */
/*
 * Copyright (c) 2010,2011,2012,2013 TELEMATICS LAB, Politecnico di Bari
 *
 * This file is part of LTE-Sim
 *
 * LTE-Sim is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 3 as
 * published by the Free Software Foundation;
 *
 * LTE-Sim is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with LTE-Sim; if not, see <http://www.gnu.org/licenses/>.
 *
 * 
 * Name: Fuzzy Scheduler
 *
 * Author: @rafaelf, @roberto
 *
 * Version: 1.5
 *
 */

#include <stdio.h>
#include <stdlib.h> 
#include <string> 
#include <math.h>

#include "fl/Headers.h"

#include "dl-fuzzy-packet-scheduler.h"
#include "../mac-entity.h"
#include "../../packet/Packet.h"
#include "../../packet/packet-burst.h"
#include "../../../device/NetworkNode.h"
#include "../../../flows/radio-bearer.h"
#include "../../../protocolStack/rrc/rrc-entity.h"
#include "../../../flows/application/Application.h"
#include "../../../device/ENodeB.h"
#include "../../../protocolStack/mac/AMCModule.h"
#include "../../../phy/lte-phy.h"
#include "../../../core/spectrum/bandwidth-manager.h"
#include "../../../flows/MacQueue.h"
#include "../../../utility/eesm-effective-sinr.h"

double coefBForCQIIndex[15] = {
  -6.816, -4.828, -2.288, -0.4206, 2.115,
  4.828, 6.06, 8.565, 10.75, 12.20,
  14.25, 16.4, 18.5, 21.35, 24.86
};

double coefCForCQIIndex[15] = {
  3.207, 3.258, 3.367, 3.462, 3.62,
  3.801, 3.552, 3.476, 3.696, 3.632,
  3.650, 3.7, 4.1, 4.407, 5.1
};

DlFuzzyPacketScheduler::DlFuzzyPacketScheduler()
{
  DefineFuzzy();
  SetMacEntity (0);
  CreateFlowsToSchedule ();
}

DlFuzzyPacketScheduler::~DlFuzzyPacketScheduler()
{
  Destroy ();
}


void
DlFuzzyPacketScheduler::DoSchedule ()
{
#ifdef SCHEDULER_DEBUG
	std::cout << "Start DL packet scheduler for node "
			<< GetMacEntity ()->GetDevice ()->GetIDNetworkNode()<< std::endl;
#endif

  UpdateAverageTransmissionRate ();
  CheckForDLDropPackets ();
  UpdatePacketLossRate();

  SelectFlowsToSchedule ();

  if (GetFlowsToSchedule ()->size() == 0)
	{}
  else
	{
	  RBsAllocation ();
	}

  StopSchedule ();
}

double DlFuzzyPacketScheduler::GetBLER(double spectralEfficiency){
  double approxBLER, sinr;
  int cqi;

  AMCModule *amc = GetMacEntity ()->GetAmcModule ();

  cqi = amc->GetCQIFromEfficiency(spectralEfficiency);
  sinr = amc->GetSinrFromCQI (cqi);

  approxBLER = 0.5 * erfc( (sinr-coefBForCQIIndex[cqi-1])/(sqrt(2.0)*coefCForCQIIndex[cqi-1]) );

  return approxBLER;
}

void 
DlFuzzyPacketScheduler::DefineFuzzy(){
  
  /*
   * Define inputs, outputs and rules of Fuzzy scheduler
   */
  int appType = 2;
  
  engine = new fl::Engine;
  input1 = new fl::InputVariable;
  input2 = new fl::InputVariable;
  input3 = new fl::InputVariable;
  output = new fl::OutputVariable;
  rule = new fl::RuleBlock;
  
  engine->setName("FuzzyScheduler");
  
  input1->setEnabled(true);
  input1->setName("HOL");
  input1->setRange(0.000, 1.000);
  input1->addTerm(new fl::Gaussian("LOW", 0.000, 0.250));
  input1->addTerm(new fl::Gaussian("MEDIUM", 0.500, 0.150));
  input1->addTerm(new fl::Gaussian("HIGH", 1.000, 0.250));
  engine->addInputVariable(input1);

  input2->setEnabled(true);
  input2->setName("BLER");
  input2->setRange(0.000, 1.000);
  input2->addTerm(new fl::Gaussian("LOW", 0.000, 0.250));
  input2->addTerm(new fl::Gaussian("MEDIUM", 0.500, 0.150));
  input2->addTerm(new fl::Gaussian("HIGH", 1.000, 0.250));
  engine->addInputVariable(input2);

  input3->setEnabled(true);
  input3->setName("AVG");
  input3->setRange(0.000, 1.000);
  input3->addTerm(new fl::Gaussian("LOW", 0.000, 0.250));
  input3->addTerm(new fl::Gaussian("MEDIUM", 0.500, 0.150));
  input3->addTerm(new fl::Gaussian("HIGH", 1.000, 0.250));
  engine->addInputVariable(input3);
  
  output->setEnabled(true);
  output->setName("METRIC");
  output->setRange(0.000, 1.000);
 
  output->fuzzyOutput()->setAccumulation(new fl::AlgebraicSum);  
  output->setDefuzzifier(new fl::Centroid(200));
  output->setDefaultValue(fl::nan);
  output->addTerm(new fl::Triangle("VERYLOW", 0.000, 0.000, 0.200));
  output->addTerm(new fl::Triangle("LOW", 0.050, 0.250, 0.450));
  output->addTerm(new fl::Triangle("MEDIUM", 0.300, 0.500, 0.700));
  output->addTerm(new fl::Triangle("HIGH", 0.550, 0.750, 0.950));
  output->addTerm(new fl::Triangle("VERYHIGH", 0.800, 1.000, 1.000));
  engine->addOutputVariable(output);

  rule->setEnabled(true);
  rule->setName("");
  rule->setConjunction(new fl::AlgebraicProduct);
  rule->setDisjunction(new fl::AlgebraicSum);
  rule->setActivation(new fl::AlgebraicProduct);
  
  /*
   * appType defines the application type
   *  
   * 1 -> VOIP
   * 2 -> TRACE_BASED/VIDEO
   * 3 -> INF_BUFFER/BE
   * 4 -> CBR
   * 
   */
  
  if (appType == 1 || appType == 3)
  {
    rule->addRule(fl::Rule::parse("if HOL is LOW and BLER is LOW and AVG is HIGH then METRIC is VERYLOW", engine));
    rule->addRule(fl::Rule::parse("if HOL is LOW and BLER is LOW and AVG is MEDIUM then METRIC is LOW", engine));
    rule->addRule(fl::Rule::parse("if HOL is LOW and BLER is LOW and AVG is LOW then METRIC is LOW", engine));
    rule->addRule(fl::Rule::parse("if HOL is LOW and BLER is MEDIUM and AVG is HIGH then METRIC is LOW", engine));
    rule->addRule(fl::Rule::parse("if HOL is LOW and BLER is MEDIUM and AVG is MEDIUM then METRIC is LOW", engine));
    rule->addRule(fl::Rule::parse("if HOL is LOW and BLER is MEDIUM and AVG is LOW then METRIC is MEDIUM", engine));
    rule->addRule(fl::Rule::parse("if HOL is LOW and BLER is HIGH and AVG is HIGH then METRIC is LOW", engine));
    rule->addRule(fl::Rule::parse("if HOL is LOW and BLER is HIGH and AVG is MEDIUM then METRIC is MEDIUM", engine));
    rule->addRule(fl::Rule::parse("if HOL is LOW and BLER is HIGH and AVG is LOW then METRIC is MEDIUM", engine));
    rule->addRule(fl::Rule::parse("if HOL is MEDIUM and BLER is LOW and AVG is HIGH then METRIC is LOW", engine));
    rule->addRule(fl::Rule::parse("if HOL is MEDIUM and BLER is LOW and AVG is MEDIUM then METRIC is LOW", engine));
    rule->addRule(fl::Rule::parse("if HOL is MEDIUM and BLER is LOW and AVG is LOW then METRIC is MEDIUM", engine));
    rule->addRule(fl::Rule::parse("if HOL is MEDIUM and BLER is MEDIUM and AVG is HIGH then METRIC is MEDIUM", engine));
    rule->addRule(fl::Rule::parse("if HOL is MEDIUM and BLER is MEDIUM and AVG is MEDIUM then METRIC is MEDIUM", engine));
    rule->addRule(fl::Rule::parse("if HOL is MEDIUM and BLER is MEDIUM and AVG is LOW then METRIC is HIGH", engine));
    rule->addRule(fl::Rule::parse("if HOL is MEDIUM and BLER is HIGH and AVG is HIGH then METRIC is MEDIUM", engine));
    rule->addRule(fl::Rule::parse("if HOL is MEDIUM and BLER is HIGH and AVG is MEDIUM then METRIC is HIGH", engine));
    rule->addRule(fl::Rule::parse("if HOL is MEDIUM and BLER is HIGH and AVG is LOW then METRIC is HIGH", engine));
    rule->addRule(fl::Rule::parse("if HOL is HIGH and BLER is LOW and AVG is HIGH then METRIC is LOW", engine));
    rule->addRule(fl::Rule::parse("if HOL is HIGH and BLER is LOW and AVG is MEDIUM then METRIC is MEDIUM", engine));
    rule->addRule(fl::Rule::parse("if HOL is HIGH and BLER is LOW and AVG is LOW then METRIC is MEDIUM", engine));
    rule->addRule(fl::Rule::parse("if HOL is HIGH and BLER is MEDIUM and AVG is HIGH then METRIC is HIGH", engine));
    rule->addRule(fl::Rule::parse("if HOL is HIGH and BLER is MEDIUM and AVG is MEDIUM then METRIC is HIGH", engine));
    rule->addRule(fl::Rule::parse("if HOL is HIGH and BLER is MEDIUM and AVG is LOW then METRIC is VERYHIGH", engine));
    rule->addRule(fl::Rule::parse("if HOL is HIGH and BLER is HIGH and AVG is HIGH then METRIC is VERYHIGH", engine));
    rule->addRule(fl::Rule::parse("if HOL is HIGH and BLER is HIGH and AVG is MEDIUM then METRIC is VERYHIGH", engine));
    rule->addRule(fl::Rule::parse("if HOL is HIGH and BLER is HIGH and AVG is LOW then METRIC is VERYHIGH", engine));    
  }

  if(appType == 2)
  {
    rule->addRule(fl::Rule::parse("if HOL is LOW and BLER is LOW and AVG is HIGH then METRIC is VERYLOW", engine));
    rule->addRule(fl::Rule::parse("if HOL is LOW and BLER is LOW and AVG is MEDIUM then METRIC is LOW", engine));
    rule->addRule(fl::Rule::parse("if HOL is LOW and BLER is LOW and AVG is LOW then METRIC is LOW", engine));
    rule->addRule(fl::Rule::parse("if HOL is LOW and BLER is MEDIUM and AVG is HIGH then METRIC is LOW", engine));
    rule->addRule(fl::Rule::parse("if HOL is LOW and BLER is MEDIUM and AVG is MEDIUM then METRIC is LOW", engine));
    rule->addRule(fl::Rule::parse("if HOL is LOW and BLER is MEDIUM and AVG is LOW then METRIC is MEDIUM", engine));
    rule->addRule(fl::Rule::parse("if HOL is LOW and BLER is HIGH and AVG is HIGH then METRIC is LOW", engine));
    rule->addRule(fl::Rule::parse("if HOL is LOW and BLER is HIGH and AVG is MEDIUM then METRIC is LOW", engine));
    rule->addRule(fl::Rule::parse("if HOL is LOW and BLER is HIGH and AVG is LOW then METRIC is MEDIUM", engine));
    rule->addRule(fl::Rule::parse("if HOL is MEDIUM and BLER is LOW and AVG is HIGH then METRIC is LOW", engine));
    rule->addRule(fl::Rule::parse("if HOL is MEDIUM and BLER is LOW and AVG is MEDIUM then METRIC is LOW", engine));
    rule->addRule(fl::Rule::parse("if HOL is MEDIUM and BLER is LOW and AVG is LOW then METRIC is MEDIUM", engine));
    rule->addRule(fl::Rule::parse("if HOL is MEDIUM and BLER is MEDIUM and AVG is HIGH then METRIC is MEDIUM", engine));
    rule->addRule(fl::Rule::parse("if HOL is MEDIUM and BLER is MEDIUM and AVG is MEDIUM then METRIC is MEDIUM", engine));
    rule->addRule(fl::Rule::parse("if HOL is MEDIUM and BLER is MEDIUM and AVG is LOW then METRIC is HIGH", engine));
    rule->addRule(fl::Rule::parse("if HOL is MEDIUM and BLER is HIGH and AVG is HIGH then METRIC is MEDIUM", engine));
    rule->addRule(fl::Rule::parse("if HOL is MEDIUM and BLER is HIGH and AVG is MEDIUM then METRIC is MEDIUM", engine));
    rule->addRule(fl::Rule::parse("if HOL is MEDIUM and BLER is HIGH and AVG is LOW then METRIC is HIGH", engine));
    rule->addRule(fl::Rule::parse("if HOL is HIGH and BLER is LOW and AVG is HIGH then METRIC is LOW", engine));
    rule->addRule(fl::Rule::parse("if HOL is HIGH and BLER is LOW and AVG is MEDIUM then METRIC is MEDIUM", engine));
    rule->addRule(fl::Rule::parse("if HOL is HIGH and BLER is LOW and AVG is LOW then METRIC is MEDIUM", engine));
    rule->addRule(fl::Rule::parse("if HOL is HIGH and BLER is MEDIUM and AVG is HIGH then METRIC is HIGH", engine));
    rule->addRule(fl::Rule::parse("if HOL is HIGH and BLER is MEDIUM and AVG is MEDIUM then METRIC is HIGH", engine));
    rule->addRule(fl::Rule::parse("if HOL is HIGH and BLER is MEDIUM and AVG is LOW then METRIC is HIGH", engine));
    rule->addRule(fl::Rule::parse("if HOL is HIGH and BLER is HIGH and AVG is HIGH then METRIC is VERYHIGH", engine));
    rule->addRule(fl::Rule::parse("if HOL is HIGH and BLER is HIGH and AVG is MEDIUM then METRIC is VERYHIGH", engine));
    rule->addRule(fl::Rule::parse("if HOL is HIGH and BLER is HIGH and AVG is LOW then METRIC is VERYHIGH", engine));
  }

  if(appType == 4)
  {
    rule->addRule(fl::Rule::parse("if HOL is LOW and BLER is LOW and AVG is HIGH then METRIC is VERYLOW", engine));
    rule->addRule(fl::Rule::parse("if HOL is LOW and BLER is LOW and AVG is MEDIUM then METRIC is VERYLOW", engine));
    rule->addRule(fl::Rule::parse("if HOL is LOW and BLER is LOW and AVG is LOW then METRIC is LOW", engine));
    rule->addRule(fl::Rule::parse("if HOL is LOW and BLER is MEDIUM and AVG is HIGH then METRIC is LOW", engine));
    rule->addRule(fl::Rule::parse("if HOL is LOW and BLER is MEDIUM and AVG is MEDIUM then METRIC is MEDIUM", engine));
    rule->addRule(fl::Rule::parse("if HOL is LOW and BLER is MEDIUM and AVG is LOW then METRIC is MEDIUM", engine));
    rule->addRule(fl::Rule::parse("if HOL is LOW and BLER is HIGH and AVG is HIGH then METRIC is LOW", engine));
    rule->addRule(fl::Rule::parse("if HOL is LOW and BLER is HIGH and AVG is MEDIUM then METRIC is MEDIUM", engine));
    rule->addRule(fl::Rule::parse("if HOL is LOW and BLER is HIGH and AVG is LOW then METRIC is MEDIUM", engine));
    rule->addRule(fl::Rule::parse("if HOL is MEDIUM and BLER is LOW and AVG is HIGH then METRIC is LOW", engine));
    rule->addRule(fl::Rule::parse("if HOL is MEDIUM and BLER is LOW and AVG is MEDIUM then METRIC is LOW", engine));
    rule->addRule(fl::Rule::parse("if HOL is MEDIUM and BLER is LOW and AVG is LOW then METRIC is MEDIUM", engine));
    rule->addRule(fl::Rule::parse("if HOL is MEDIUM and BLER is MEDIUM and AVG is HIGH then METRIC is MEDIUM", engine));
    rule->addRule(fl::Rule::parse("if HOL is MEDIUM and BLER is MEDIUM and AVG is MEDIUM then METRIC is MEDIUM", engine));
    rule->addRule(fl::Rule::parse("if HOL is MEDIUM and BLER is MEDIUM and AVG is LOW then METRIC is HIGH", engine));
    rule->addRule(fl::Rule::parse("if HOL is MEDIUM and BLER is HIGH and AVG is HIGH then METRIC is MEDIUM", engine));
    rule->addRule(fl::Rule::parse("if HOL is MEDIUM and BLER is HIGH and AVG is MEDIUM then METRIC is MEDIUM", engine));
    rule->addRule(fl::Rule::parse("if HOL is MEDIUM and BLER is HIGH and AVG is LOW then METRIC is HIGH", engine));
    rule->addRule(fl::Rule::parse("if HOL is HIGH and BLER is LOW and AVG is HIGH then METRIC is LOW", engine));
    rule->addRule(fl::Rule::parse("if HOL is HIGH and BLER is LOW and AVG is MEDIUM then METRIC is MEDIUM", engine));
    rule->addRule(fl::Rule::parse("if HOL is HIGH and BLER is LOW and AVG is LOW then METRIC is MEDIUM", engine));
    rule->addRule(fl::Rule::parse("if HOL is HIGH and BLER is MEDIUM and AVG is HIGH then METRIC is MEDIUM", engine));
    rule->addRule(fl::Rule::parse("if HOL is HIGH and BLER is MEDIUM and AVG is MEDIUM then METRIC is HIGH", engine));
    rule->addRule(fl::Rule::parse("if HOL is HIGH and BLER is MEDIUM and AVG is LOW then METRIC is HIGH", engine));
    rule->addRule(fl::Rule::parse("if HOL is HIGH and BLER is HIGH and AVG is HIGH then METRIC is VERYHIGH", engine));
    rule->addRule(fl::Rule::parse("if HOL is HIGH and BLER is HIGH and AVG is MEDIUM then METRIC is VERYHIGH", engine));
    rule->addRule(fl::Rule::parse("if HOL is HIGH and BLER is HIGH and AVG is LOW then METRIC is VERYHIGH", engine));
  }
  engine->addRuleBlock(rule);
  
}

double
DlFuzzyPacketScheduler::ComputeSchedulingMetric (RadioBearer *bearer, double spectralEfficiency, int subChannel)
{

  /*
   * For the Fuzzy scheduler the metric is computed
   * as follows:
   *
   * metric = defuzzify(bler, hol, avg)
   * 
   *
   * where:
   *
   * bler = (tanh(xbler-1)+1)/2; xbler = measuredBler/blerThreshold
   * hol = (tanh(xhol-1)+1)/2; xhol = measuredHol/holThreshold
   * avg = averageRate/maxBandwidth
   * 
   */

  // Thresholds
  double blerThreshold = 0.1075;
  double holThreshold = 0.1;
  
  // Inputs
  double bler = tanh(GetBLER(spectralEfficiency)/blerThreshold); 
  double hol = tanh(bearer->GetHeadOfLinePacketDelay ()/holThreshold);
  double avg = tanh(bearer->GetAverageTransmissionRate ()/5000000.0); 
  
//  double bler = (1/2)*(tanh((approxBLER/blerThreshold)-1)+1); 
//  double hol = (1/2)*(tanh((bearer->GetHeadOfLinePacketDelay ()/holThreshold)-1)+1);
//  double avg = bearer->GetAverageTransmissionRate ()/25000000.0; 

  input1->setInputValue(bler);
  input2->setInputValue(hol);
  input3->setInputValue(avg);
  
  engine->process();

  // Compute metric
  double metric = engine->getOutputValue("METRIC");

  return metric;
}
