/* -*- Mode:C++; c-file-style:"gnu"; indent-tabs-mode:nil; -*- */
/*
 * Copyright (c) 2010,2011,2012,2013 TELEMATICS LAB, Politecnico di Bari
 *
 * This file is part of LTE-Sim
 *
 * LTE-Sim is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 3 as
 * published by the Free Software Foundation;
 *
 * LTE-Sim is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with LTE-Sim; if not, see <http://www.gnu.org/licenses/>.
 *
 * 
 * Name: Fuzzy Scheduler
 *
 * Author: @rafaelf, @roberto
 *
 * Version: 1.5
 *
 */

#include <stdio.h>
#include <stdlib.h> 
#include <string> 
#include <math.h>

#include "fl/Headers.h"

#include "dl-fuzzy-packet-scheduler.h"
#include "../mac-entity.h"
#include "../../packet/Packet.h"
#include "../../packet/packet-burst.h"
#include "../../../device/NetworkNode.h"
#include "../../../flows/radio-bearer.h"
#include "../../../protocolStack/rrc/rrc-entity.h"
#include "../../../flows/application/Application.h"
#include "../../../device/ENodeB.h"
#include "../../../protocolStack/mac/AMCModule.h"
#include "../../../phy/lte-phy.h"
#include "../../../core/spectrum/bandwidth-manager.h"
#include "../../../flows/MacQueue.h"
#include "../../../utility/eesm-effective-sinr.h"


DlFuzzyPacketScheduler::DlFuzzyPacketScheduler()
{
  DefineFuzzy();
  SetMacEntity (0);
  CreateFlowsToSchedule ();
}

DlFuzzyPacketScheduler::~DlFuzzyPacketScheduler()
{
  Destroy ();
}


void
DlFuzzyPacketScheduler::DoSchedule ()
{
#ifdef SCHEDULER_DEBUG
	std::cout << "Start DL packet scheduler for node "
			<< GetMacEntity ()->GetDevice ()->GetIDNetworkNode()<< std::endl;
#endif

  UpdateAverageTransmissionRate ();
  CheckForDLDropPackets ();
  UpdatePacketLossRate();

  SelectFlowsToSchedule ();

  if (GetFlowsToSchedule ()->size() == 0)
	{}
  else
	{
	  RBsAllocation ();
	}

  StopSchedule ();
}

void 
DlFuzzyPacketScheduler::DefineFuzzy(){
  
  /*
   * Define inputs, outputs and rules of Fuzzy scheduler
   *
   * it >> Instant Throughput
   * hol >> Head-Of-Line Packet Delay
   * avg >> Average Rate
   */
  
  engine = new fl::Engine;
  input1 = new fl::InputVariable;
  input2 = new fl::InputVariable;
  output = new fl::OutputVariable;
  rule = new fl::RuleBlock;
  
  engine->setName("FuzzyScheduler");
  
  input1->setEnabled(true);
  input1->setName("IT");
  input1->setRange(0.000, 1.000);
  input1->addTerm(new fl::Gaussian("LOW", 0.000, 0.250));
  input1->addTerm(new fl::Gaussian("MEDIUM", 0.500, 0.150));
  input1->addTerm(new fl::Gaussian("HIGH", 1.000, 0.250));
  engine->addInputVariable(input1);

  input2->setEnabled(true);
  input2->setName("AVG");
  input2->setRange(0.000, 1.000);
  input2->addTerm(new fl::Gaussian("LOW", 0.000, 0.250));
  input2->addTerm(new fl::Gaussian("MEDIUM", 0.500, 0.150));
  input2->addTerm(new fl::Gaussian("HIGH", 1.000, 0.250));
  engine->addInputVariable(input2);

  output->setEnabled(true);
  output->setName("METRIC");
  output->setRange(0.000, 1.000);
 
  output->fuzzyOutput()->setAccumulation(new fl::AlgebraicSum);  
  output->setDefuzzifier(new fl::Centroid(200));
  output->setDefaultValue(fl::nan);
  output->addTerm(new fl::Triangle("VERYLOW", 0.000, 0.000, 0.200));
  output->addTerm(new fl::Triangle("LOW", 0.050, 0.250, 0.450));
  output->addTerm(new fl::Triangle("MEDIUM", 0.300, 0.500, 0.700));
  output->addTerm(new fl::Triangle("HIGH", 0.550, 0.750, 0.950));
  output->addTerm(new fl::Triangle("VERYHIGH", 0.800, 1.000, 1.000));
  engine->addOutputVariable(output);

  rule->setEnabled(true);
  rule->setName("");
  rule->setConjunction(new fl::AlgebraicProduct);
  rule->setDisjunction(new fl::AlgebraicSum);
  rule->setActivation(new fl::AlgebraicProduct);
  
  rule->addRule(fl::Rule::parse("if IT is HIGH and AVG is HIGH then METRIC is VERYLOW", engine));
  rule->addRule(fl::Rule::parse("if IT is HIGH and AVG is MEDIUM then METRIC is LOW", engine));
  rule->addRule(fl::Rule::parse("if IT is HIGH and AVG is LOW then METRIC is MEDIUM", engine));
  rule->addRule(fl::Rule::parse("if IT is MEDIUM and AVG is HIGH then METRIC is LOW", engine));
  rule->addRule(fl::Rule::parse("if IT is MEDIUM and AVG is MEDIUM then METRIC is MEDIUM", engine));
  rule->addRule(fl::Rule::parse("if IT is MEDIUM and AVG is LOW then METRIC is HIGH", engine));
  rule->addRule(fl::Rule::parse("if IT is LOW and AVG is HIGH then METRIC is MEDIUM", engine));
  rule->addRule(fl::Rule::parse("if IT is LOW and AVG is MEDIUM then METRIC is HIGH", engine));
  rule->addRule(fl::Rule::parse("if IT is LOW and AVG is LOW then METRIC is VERYHIGH", engine));


  engine->addRuleBlock(rule);
  
}

double
DlFuzzyPacketScheduler::ComputeSchedulingMetric (RadioBearer *bearer, double spectralEfficiency, int subChannel)
{

  /*
   * For the Fuzzy scheduler the metric is computed
   * as follows:
   *
   * metric = defuzzify(it, hol, avg)
   * 
   *
   * where:
   *
   * it = (tanh(xit-1)+1)/2; xit = measuredIt/itThreshold
   * hol = (tanh(xhol-1)+1)/2; xhol = measuredHol/holThreshold
   * avg = averageRate/maxBandwidth
   * 
   */

  // Thresholds
  double itThreshold = 500000.;
  double maxBandwidth = 1000000.;
  
  // Inputs
  double xit = spectralEfficiency * 180000./itThreshold;
  double xavg = bearer->GetAverageTransmissionRate()/maxBandwidth;

  double it = tanh(xit); //(1/2)*(tanh(xit-1)+1);
  double avg = tanh(xavg); //(1/2)*(tanh(xavg-1)+1);

  input1->setInputValue(it);
  input2->setInputValue(avg);
  
  engine->process();

  // Compute metric
  double metric = engine->getOutputValue("METRIC");
  

  std::cout << "xit(" << spectralEfficiency * 180000. << "/" << itThreshold << "): " << xit << " @ yit: " << it << 
     //     " -- xhol(" << bearer->GetHeadOfLinePacketDelay() << "/" << holThreshold << "): " << xhol << " @ yhol: " << hol << 
          " -- xavg(" << bearer->GetAverageTransmissionRate() << "/" << maxBandwidth << "): " << xavg << " @ yavg: " << avg << 
          " >>>>>>>>> metric: " << metric << 
  std::endl;


  return metric;
}
