/* -*- Mode:C++; c-file-style:"gnu"; indent-tabs-mode:nil; -*- */
/*
 * Copyright (c) 2010,2011,2012,2013 TELEMATICS LAB, Politecnico di Bari
 *
 * This file is part of LTE-Sim
 *
 * LTE-Sim is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 3 as
 * published by the Free Software Foundation;
 *
 * LTE-Sim is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
		 * along with LTE-Sim; if not, see <http://www.gnu.org/licenses/>.
		 *
 * 
 * Name: Fuzzy Scheduler
 *
 * Author: @rafaelf, @roberto
 *
 * Version: 2.0
 *
 */

#include <stdio.h>
#include <stdlib.h> 
#include <string> 
#include <math.h>
#include <float.h>

#include "fl/Headers.h"

#include "dl-fuzzy-packet-scheduler.h"
#include "../mac-entity.h"
#include "../../packet/Packet.h"
#include "../../packet/packet-burst.h"
#include "../../../device/NetworkNode.h"
#include "../../../flows/radio-bearer.h"
#include "../../../protocolStack/rrc/rrc-entity.h"
#include "../../../flows/application/Application.h"
#include "../../../device/ENodeB.h"
#include "../../../protocolStack/mac/AMCModule.h"
#include "../../../phy/lte-phy.h"
#include "../../../core/spectrum/bandwidth-manager.h"
#include "../../../flows/MacQueue.h"
#include "../../../flows/QoS/QoSForFuzzy.h"
#include "../../../utility/eesm-effective-sinr.h"

double coefBForCQIIndex[15] = {
  -6.816, -4.828, -2.288, -0.4206, 2.115,
  4.828, 6.06, 8.565, 10.75, 12.20,
  14.25, 16.4, 18.5, 21.35, 24.86
};

double coefCForCQIIndex[15] = {
  3.207, 3.258, 3.367, 3.462, 3.62,
  3.801, 3.552, 3.476, 3.696, 3.632,
  3.650, 3.7, 4.1, 4.407, 5.1
};

DlFuzzyPacketScheduler::DlFuzzyPacketScheduler()
{
  DefineFuzzy();
  SetMacEntity (0);
  CreateFlowsToSchedule ();
}

DlFuzzyPacketScheduler::~DlFuzzyPacketScheduler()
{
  Destroy ();
}


void
DlFuzzyPacketScheduler::DoSchedule ()
{
#ifdef SCHEDULER_DEBUG
	std::cout << "Start DL packet scheduler for node "
			<< GetMacEntity ()->GetDevice ()->GetIDNetworkNode()<< std::endl;
#endif

  UpdateAverageTransmissionRate ();
  CheckForDLDropPackets ();
  UpdatePacketLossRate();

  SelectFlowsToSchedule ();

  if (GetFlowsToSchedule ()->size() == 0)
	{}
  else
	{
	  RBsAllocation ();
	}

  StopSchedule ();
}

double DlFuzzyPacketScheduler::GetBLER(double spectralEfficiency){
  double approxBLER, sinr;
  int cqi;

  AMCModule *amc = GetMacEntity ()->GetAmcModule ();

  cqi = amc->GetCQIFromEfficiency(spectralEfficiency);
  sinr = amc->GetSinrFromCQI (cqi);

  approxBLER = 0.5 * erfc( (sinr-coefBForCQIIndex[cqi-1])/(sqrt(2.0)*coefCForCQIIndex[cqi-1]) );

  return approxBLER;
}

void DlFuzzyPacketScheduler::DefineFuzzy(){
  
  /*
   * Define inputs, outputs and rules of Fuzzy scheduler
   *
   * hol >> Head-Of-Line Packet Delay
   * it  >> Instant Throughput
   * avg >> Average Rate
   */
  
  engine = new fl::Engine;
  input1 = new fl::InputVariable;
  input2 = new fl::InputVariable;
  input3 = new fl::InputVariable;
  input4 = new fl::InputVariable;
  output = new fl::OutputVariable;
  rule = new fl::RuleBlock;
  
  engine->setName("FuzzyScheduler");
  
  input1->setEnabled(true);
  input1->setName("HOL");
  input1->setRange(0.000, 1.000);
  input1->addTerm(new fl::Triangle("LOW", 0.000, 0.000, 0.800));
  input1->addTerm(new fl::Triangle("HIGH", 0.200, 1.000, 1.000));
  engine->addInputVariable(input1);

  input2->setEnabled(true);
  input2->setName("IT");
  input2->setRange(0.000, 1.000);
  input2->addTerm(new fl::Triangle("LOW", 0.000, 0.000, 0.800));
  input2->addTerm(new fl::Triangle("HIGH", 0.200, 1.000, 1.000));
  engine->addInputVariable(input2);

  input3->setEnabled(true);
  input3->setName("AVG");
  input3->setRange(0.000, 1.000);
  input3->addTerm(new fl::Triangle("LOW", 0.000, 0.000, 0.800));
  input3->addTerm(new fl::Triangle("HIGH", 0.200, 1.000, 1.000));
  engine->addInputVariable(input3);
 /*
  input4->setEnabled(true);
  input4->setName("BLER");
  input4->setRange(0.000, 1.000);
  input4->addTerm(new fl::Triangle("LOW", 0.000, 0.350, 0.700));
  input4->addTerm(new fl::Triangle("HIGH", 0.350, 0.700, 1.000));
  engine->addInputVariable(input4);
*/
  output->setEnabled(true);
  output->setName("METRIC");
  output->setRange(0.000, 1.000);
 
  output->fuzzyOutput()->setAccumulation(new fl::AlgebraicSum);  
  output->setDefuzzifier(new fl::Centroid(200));
  output->setDefaultValue(fl::nan);
  output->addTerm(new fl::Triangle("VERYLOW", 0.000, 0.000, 0.200));
  output->addTerm(new fl::Triangle("LOW", 0.050, 0.250, 0.450));
  output->addTerm(new fl::Triangle("MEDIUM", 0.300, 0.500, 0.700));
  output->addTerm(new fl::Triangle("HIGH", 0.550, 0.750, 0.950));
  output->addTerm(new fl::Triangle("VERYHIGH", 0.800, 1.000, 1.000));
  engine->addOutputVariable(output);

  rule->setEnabled(true);
  rule->setName("");
  rule->setConjunction(new fl::AlgebraicProduct);
  rule->setDisjunction(new fl::AlgebraicSum);
  rule->setActivation(new fl::AlgebraicProduct);
  
  rule->addRule(fl::Rule::parse("if IT is LOW and AVG is HIGH and HOL is LOW then METRIC is VERYLOW", engine));
  rule->addRule(fl::Rule::parse("if IT is LOW and AVG is HIGH and HOL is HIGH then METRIC is LOW", engine));
  rule->addRule(fl::Rule::parse("if IT is LOW and AVG is LOW and HOL is LOW then METRIC is MEDIUM", engine));
  rule->addRule(fl::Rule::parse("if IT is LOW and AVG is LOW and HOL is HIGH then METRIC is MEDIUM", engine));
  rule->addRule(fl::Rule::parse("if IT is HIGH and AVG is HIGH and HOL is LOW then METRIC is HIGH", engine));
  rule->addRule(fl::Rule::parse("if IT is HIGH and AVG is HIGH and HOL is HIGH then METRIC is HIGH", engine));
  rule->addRule(fl::Rule::parse("if IT is HIGH and AVG is LOW and HOL is LOW then METRIC is VERYHIGH", engine));
  rule->addRule(fl::Rule::parse("if IT is HIGH and AVG is LOW and HOL is HIGH then METRIC is VERYHIGH", engine));
   
  engine->addRuleBlock(rule);
  
}

double DlFuzzyPacketScheduler::ComputeSchedulingMetric (RadioBearer *bearer, double spectralEfficiency, int subChannel)
{
  return 0;
}

double
DlFuzzyPacketScheduler::ComputeSchedulingMetric (RadioBearer *bearer, double spectralEfficiency, int subChannel, double maxBLER, double maxInstantThroughput, double maxAverageThroughput, double minBLER, double minInstantThroughput, double minAverageThroughput)
{

  /*
   * For the Fuzzy scheduler the metric is computed
   * as follows:
   *
   * metric = defuzzify(it, hol, avg)
   * 
   *
   * where:
   *
   * it = (tanh(xit-1)+1)/2; xit = measuredIt/itThreshold
   * hol = (tanh(xhol-1)+1)/2; xhol = measuredHol/holThreshold
   * avg = averageRate/maxBandwidth
   * 
   */

   QoSForFuzzy *qos = (QoSForFuzzy*) bearer->GetQoSParameters ();

  // Thresholds
  double holThreshold = qos->GetMaxDelay ();
  
  // Inputs
  double xhol = bearer->GetHeadOfLinePacketDelay()/holThreshold;
  
  double hol = tanh(xhol);
  double it = (spectralEfficiency-minInstantThroughput)/(maxInstantThroughput-minInstantThroughput);
  double avg = (bearer->GetAverageTransmissionRate()-minAverageThroughput)/(maxAverageThroughput-minAverageThroughput);
 // double bler = (GetBLER(spectralEfficiency)-minBLER)/(maxBLER-minBLER); 

  if(!(it >= 0 && it <= 1)){
    it = 0.005;
  }
  if(!(avg >= 0 && avg <= 1)){
    avg = 0.005;
  }
 /* if(!(bler >= 0 && bler <= 1)){
    bler = 0.5;
  } 
*/
  if(hol == 0) hol = 0.005;
  if(hol == 1) hol = 0.995;

  if(it == 0) it = 0.005;
  if(it == 1) it = 0.995;
 
  if(avg == 0) avg = 0.005;
  if(avg == 1) avg = 0.995;
/*
  if(bler == 0) bler = 0.005;
  if(bler == 1) bler = 0.995; 
*/
  input1->setInputValue(hol);
  input2->setInputValue(it);
  input3->setInputValue(avg);
 // input4->setInputValue(bler); 
 
  engine->process();

  // Compute metric
  double metric = engine->getOutputValue("METRIC");

/*
  printf("hol(tanh(%.2f/%.2f)) -> %.5f\n", bearer->GetHeadOfLinePacketDelay(), holThreshold, hol); 
  printf("it(%.2f/%.2f) -> %.5f\n", (spectralEfficiency-minInstantThroughput), (maxInstantThroughput-minInstantThroughput), it);
  printf("avg(%.2f/%.2f) -> %.5f\n", (bearer->GetAverageTransmissionRate()-minAverageThroughput), (maxAverageThroughput-minAverageThroughput), avg);
 // printf("bler(%.3f/%.3f) -> %.5f\n", (GetBLER(spectralEfficiency)-minBLER), (maxBLER-minBLER), bler);  
  printf("plr -> %.5f\n", bearer->GetMacQueue()->GetPacketLossRate());
  printf("metric -> %.5f\n\n", metric);
*/

  return metric;
}

void
DlFuzzyPacketScheduler::RBsAllocation ()
{
#ifdef SCHEDULER_DEBUG
	std::cout << " ---- DownlinkPacketScheduler::RBsAllocation";
#endif

  FlowsToSchedule* flows = GetFlowsToSchedule ();
  int nbOfRBs = GetMacEntity ()->GetDevice ()->GetPhy ()->GetBandwidthManager ()->GetDlSubChannels ().size ();
 
  //create a matrix of flow metrics
  double metrics[nbOfRBs][flows->size ()];
  for (int i = 0; i < nbOfRBs; i++)
    {

	  double maxBLER = 0;
	  double minBLER = DBL_MAX; //100000000;
	  double maxInstantThroughput = 0;
	  double maxAverageThroughput = 0;
	  double minInstantThroughput = DBL_MAX; //100000000;
	  double minAverageThroughput = DBL_MAX; //100000000;

	  for (int j = 0; j < flows->size (); j++)
	    {
		// maxs
		  if(GetBLER(flows->at (j)->GetSpectralEfficiency ().at (i)) > maxBLER){
                          maxBLER = GetBLER(flows->at (j)->GetSpectralEfficiency ().at (i));
                   }

  		  if(flows->at (j)->GetSpectralEfficiency ().at (i) > maxInstantThroughput){
		          maxInstantThroughput = flows->at (j)->GetSpectralEfficiency ().at (i);		  
		   }
		  
		  if(flows->at (j)->GetBearer ()->GetAverageTransmissionRate() > maxAverageThroughput){
			  maxAverageThroughput = flows->at (j)->GetBearer ()->GetAverageTransmissionRate();
		  }


		// mins
		 if(GetBLER(flows->at (j)->GetSpectralEfficiency ().at (i)) < minBLER){
                          minBLER = GetBLER(flows->at (j)->GetSpectralEfficiency ().at (i));
                   }

                  if(flows->at (j)->GetSpectralEfficiency ().at (i) < minInstantThroughput){
                          minInstantThroughput = flows->at (j)->GetSpectralEfficiency ().at (i);
                   }

                  if(flows->at (j)->GetBearer ()->GetAverageTransmissionRate() < minAverageThroughput){
                          minAverageThroughput = flows->at (j)->GetBearer ()->GetAverageTransmissionRate();
                  }

		}


		for (int j = 0; j < flows->size (); j++)
            	{
		  metrics[i][j] = ComputeSchedulingMetric (flows->at (j)->GetBearer (), flows->at (j)->GetSpectralEfficiency ().at (i), i, maxBLER, maxInstantThroughput, maxAverageThroughput, minBLER, minInstantThroughput, minAverageThroughput);
	    	}

    }

#ifdef SCHEDULER_DEBUG
  std::cout << ", available RBs " << nbOfRBs << ", flows " << flows->size () << std::endl;
  for (int ii = 0; ii < flows->size (); ii++)
    {
	  std::cout << "\t metrics for flow "
			  << flows->at (ii)->GetBearer ()->GetApplication ()->GetApplicationID () << ":";
	  for (int jj = 0; jj < nbOfRBs; jj++)
	    {
		  std::cout << " " << metrics[jj][ii];
	    }
	  std::cout << std::endl;
    }
#endif


  AMCModule *amc = GetMacEntity ()->GetAmcModule ();
  double l_dAllocatedRBCounter = 0;

  int l_iNumberOfUsers = ((ENodeB*)this->GetMacEntity()->GetDevice())->GetNbOfUserEquipmentRecords();

  bool * l_bFlowScheduled = new bool[flows->size ()];
  int l_iScheduledFlows = 0;
  std::vector<double> * l_bFlowScheduledSINR = new std::vector<double>[flows->size ()];
  for (int k = 0; k < flows->size (); k++)
      l_bFlowScheduled[k] = false;

  //RBs allocation
  for (int s = 0; s < nbOfRBs; s++)
    {
      if (l_iScheduledFlows == flows->size ())
          break;

      double targetMetric = 0;
      bool RBIsAllocated = false;
      FlowToSchedule* scheduledFlow;
      int l_iScheduledFlowIndex = 0;

      for (int k = 0; k < flows->size (); k++)
        {
          if (metrics[s][k] > targetMetric && !l_bFlowScheduled[k])
            {
              targetMetric = metrics[s][k];
              RBIsAllocated = true;
              scheduledFlow = flows->at (k);
              l_iScheduledFlowIndex = k;
            }
        }

      if (RBIsAllocated)
        {
          l_dAllocatedRBCounter++;

          scheduledFlow->GetListOfAllocatedRBs()->push_back (s); // the s RB has been allocated to that flow!

#ifdef SCHEDULER_DEBUG
          std::cout << "\t *** RB " << s << " assigned to the "
                  " flow " << scheduledFlow->GetBearer ()->GetApplication ()->GetApplicationID ()
                  << std::endl;
#endif
          double sinr = amc->GetSinrFromCQI (scheduledFlow->GetCqiFeedbacks ().at (s));
          l_bFlowScheduledSINR[l_iScheduledFlowIndex].push_back(sinr);

          double effectiveSinr = GetEesmEffectiveSinr (l_bFlowScheduledSINR[l_iScheduledFlowIndex]);
          int mcs = amc->GetMCSFromCQI (amc->GetCQIFromSinr (effectiveSinr));
          int transportBlockSize = amc->GetTBSizeFromMCS (mcs, scheduledFlow->GetListOfAllocatedRBs ()->size ());
          if (transportBlockSize >= scheduledFlow->GetDataToTransmit() * 8)
          {
              l_bFlowScheduled[l_iScheduledFlowIndex] = true;
              l_iScheduledFlows++;
          }

        }
    }

  delete [] l_bFlowScheduled;
  delete [] l_bFlowScheduledSINR;


  //Finalize the allocation
  PdcchMapIdealControlMessage *pdcchMsg = new PdcchMapIdealControlMessage ();

  for (FlowsToSchedule::iterator it = flows->begin (); it != flows->end (); it++)
    {
      FlowToSchedule *flow = (*it);
      if (flow->GetListOfAllocatedRBs ()->size () > 0)
        {
          //this flow has been scheduled
          std::vector<double> estimatedSinrValues;
          for (int rb = 0; rb < flow->GetListOfAllocatedRBs ()->size (); rb++ )

            {
              double sinr = amc->GetSinrFromCQI (
                      flow->GetCqiFeedbacks ().at (flow->GetListOfAllocatedRBs ()->at (rb)));

              estimatedSinrValues.push_back (sinr);
            }

          //compute the effective sinr
          double effectiveSinr = GetEesmEffectiveSinr (estimatedSinrValues);

          //get the MCS for transmission

          int mcs = amc->GetMCSFromCQI (amc->GetCQIFromSinr (effectiveSinr));

          //define the amount of bytes to transmit
          //int transportBlockSize = amc->GetTBSizeFromMCS (mcs);
          int transportBlockSize = amc->GetTBSizeFromMCS (mcs, flow->GetListOfAllocatedRBs ()->size ());
          double bitsToTransmit = transportBlockSize;
          flow->UpdateAllocatedBits (bitsToTransmit);

#ifdef SCHEDULER_DEBUG
		  std::cout << "\t\t --> flow "	<< flow->GetBearer ()->GetApplication ()->GetApplicationID ()
				  << " has been scheduled: " <<
				  "\n\t\t\t nb of RBs " << flow->GetListOfAllocatedRBs ()->size () <<
				  "\n\t\t\t effectiveSinr " << effectiveSinr <<
				  "\n\t\t\t tbs " << transportBlockSize <<
				  "\n\t\t\t bitsToTransmit " << bitsToTransmit
				  << std::endl;
#endif

		  //create PDCCH messages
		  for (int rb = 0; rb < flow->GetListOfAllocatedRBs ()->size (); rb++ )
		    {
			  pdcchMsg->AddNewRecord (PdcchMapIdealControlMessage::DOWNLINK,
					  flow->GetListOfAllocatedRBs ()->at (rb),
									  flow->GetBearer ()->GetDestination (),
									  mcs);
		    }
	    }
    }

  if (pdcchMsg->GetMessage()->size () > 0)
    {
      GetMacEntity ()->GetDevice ()->GetPhy ()->SendIdealControlMessage (pdcchMsg);
    }
  delete pdcchMsg;
}
